#include <errno.h> /* errno */
#include <fcntl.h> /* open(2) */
#include <stdio.h> /* fprintf(3), stderr */
#include <stdlib.h> /* free(3), malloc(3), strtol(3), size_t */
#include <string.h> /* memcpy(3), memset(3) */
#include <sysexits.h> /* EX_OK, EX_USAGE */
#include <unistd.h> /* close(2), getopt(3), lseek(2), read(2), write(2),
                       optarg, optind, STDIN_FILENO, STDOUT_FILENO */

/* Macro to use the appropriate string for argv[0] in the case that
 * argc == 0. */
#define ARGV0 (argv[0] == NULL ? "<no argv[0]>" : argv[0])

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

/* Macro to ensure fd is not a std* file, e.g. stdin. */
#define fdisnotstd(fd) \
		((fd) != STDIN_FILENO \
		&& (fd) != STDOUT_FILENO \
		&& (fd) != STDERR_FILENO)

/* Macro to call the cleanup functions that operate on struct io on the
 * particular io[2] used in main. */
#define terminate(io) do{ \
	Io_buffree(&(io)[0]); \
	Io_buffree(&(io)[1]); \
	Io_fdclose(&(io)[0]); \
	Io_fdclose(&(io)[1]); }while(0)

static int read_flags = O_RDONLY;
static int write_flags = O_WRONLY | O_CREAT;
static char *stdin_name = "<stdin>";
static char *stdout_name = "<stdout>";

/* information that needs to be kept about each file in the "pipe" */
struct Io{
	int bs;			/* buffer size */
	size_t bufuse;	/* buffer usage */
	char *buf;		/* buffer */
	int fd;			/* file descriptor */
	int fl;			/* file opening flags */
	char *fn;		/* file name (may be stdin_name or stdout_name) */
	int prec;		/* partial records processed */
	int rec;		/* records processed */
	int seek;		/* bytes to seek/skip (will be 0 after skippage) */
};

/* Allocates *io's buffer. Returns NULL if successful. */
static void *
Io_bufalloc(struct Io *io){

	return (io->buf = malloc(io->bs * (sizeof io->buf)));
}

/* Frees *io's buffer. Returns io. */
static struct Io *
Io_buffree(struct Io *io){

	free(io->buf);

	return io;
}

/* Closes io->fn and returns -1 on error, otherwise io->fd. */
static int
Io_fdclose(struct Io *io){

	return fdisnotstd(io->fd)
		? close(io->fd)
		: 0;
}

/* Opens io->fn and saves the file descriptor into io->fd. Returns io->fd,
 * which will be -1 if an error occured. */
static int
Io_fdopen(struct Io *io, char *fn){
	int fd;
	
	if((fd = open(fn, io->fl)) != -1
			&& (fdisnotstd(io->fd) || close(io->fd) == 0)){
		io->fd = fd;
		io->fn = fn;
	}

	return fd;
}

/* Reads io->bs bytes from *io's file descriptor into io->buf, storing the
 * number of read bytes in io->bufuse. If io->bufuse is 0, errno will probably
 * be set. Returns io. */
static struct Io *
Io_read(struct Io *io){

	io->bufuse = read(io->fd, io->buf, io->bs);

	return io;
}

/* Seeks io->seek bytes through *io's file descriptor, (counter-intuitively)
 * returning -1 if successful and a sysexits.h exit code if an unrecoverable
 * error occurred. */
static int
Io_seek(struct Io *io){
	
	if(fdisnotstd(io->fd) && lseek(io->fd, io->seek, SEEK_SET) == -1)
		return -1;

	memset(io->buf, '\0', io->bs);

	do{	if(		(io->fl == read_flags
					&& (((io->bufuse = read(io->fd, io->buf,
							MIN(io->bs, io->seek))) != 0)
						/* second chance if 0 */
						|| ((io->bufuse = read(io->fd, io->buf,
							MIN(io->bs, io->seek))) != 0)))
				/* cheat and use bufuse as the retval for write
				 * (the buffer is zeroed, it doesn't matter) */
				|| (io->fl == write_flags
					&& ((io->bufuse = write(io->fd, io->buf,
							MIN(io->bs, io->seek))) != 0)))
			io->seek -= io->bufuse;
		else
			return EX_OK;
	}while(io->seek > 0);

	return -1;
}

/* Sets the variables in a struct *io to the defaults. Identifies the read/
 * write ends of the "pipe" by checking io->fl. Returns io. */
static struct Io *
Io_setdefaults(struct Io *io){

	io->bs = 1024;
	io->buf = NULL;
	io->fd = (io->fl == read_flags) ? STDIN_FILENO : STDOUT_FILENO;
	io->fn = (io->fl == read_flags) ? stdin_name   : stdout_name;
	io->prec = 0;
	io->rec = 0;
	io->seek = 0;

	return io;
}

/* Parses the string s to an integer, returning either the integer or in the
 * case of an error a negative integer. This is used for argument parsing
 * (e.g. -B [int]) in dj and no negative integer would be valid anyway. */
static int
parse(char *s){
	int r;

	errno = 0;
	r = (int)strtol(s, &s, 0);
	return (*s == '\0' /* empty string */ || errno != 0)
		? -1
		: r;
}

/* Prints an error message suitable for the event of an operating system error,
 * with the error itself to be described in the string s. */
static int
oserr(char *argv0, char *s){

	fprintf(stderr, "%s: %s: %s\n", argv0, s, strerror(errno));

	return EX_OSERR;
}

/* Prints statistics regarding the use of dj, particularly partially and
 * completely read and written records. */
struct Io *
output(struct Io io[2]){

	fprintf(stderr, "%d:%d / %d:%d\n", /* dd(1p) does %d+%d\n%d+%d\n */
		io[0].rec, io[0].prec, io[1].rec, io[1].prec);;

	return io;
}

static int
usage(char **argv){

	fprintf(stderr, "Usage: %s (-fA) (-a [byte]) (-c [count]\n"
		"\t(-i [input file]) (-b [input block size])"
			" (-s [input offset])\n"
		"\t(-o [output file]) (-B [output block size])"
			" (-S [output offset])\n", ARGV0);

	return EX_USAGE;
}


int main(int argc, char *argv[]){
	char align;
	char noerror;
	int c;
	int count;
	int p; /* pipe side */
	struct Io io[2]; /* { read, write } same as pipe(2) */

	/* defaults */
	align = -1;
	count = 0;
	noerror = 0;
	io[0].fl = read_flags;
	Io_setdefaults(&io[0]);
	io[1].fl = write_flags;
	Io_setdefaults(&io[1]);

	if(argc > 0)
		while((c = getopt(argc, argv, "aAc:i:b:fs:o:B:S:")) != -1)
			if((c == 'i' || c == 'o')
					&& Io_fdopen(&io[c == 'o'], optarg) != 0){ /* file args */
				terminate(io);
				return oserr(argv[0], optarg);
			/* boolean/flag arguments */
			}else if(c == 'f')
				noerror = 1;
			else if(c == 'A')
				align = '\0';
			/* optarg */
			else if(c == 'a' && optarg[0] != '\0' && optarg[1] == '\0')
				align = optarg[0];
			/* numeric args as a double negative to catch uncaught stuff */
			else if(!((c == 'c' && (count = parse(optarg)) < 0)
					|| (c == 'b' && (io[0].bs = parse(optarg)) < 0)
					|| (c == 's' && (io[0].seek = parse(optarg)) < 0)
					|| (c == 'B' && (io[1].bs = parse(optarg)) < 0)
					|| (c == 'S' && (io[1].seek = parse(optarg)) < 0))){
				terminate(io);
				return usage(argv);
			}

	if(argc > optind)
		return usage(argv);

	for(p = 0; p <= 1; ++p){
		if(Io_bufalloc(&io[p]) == NULL){
			fprintf(stderr, "%s: Failed to allocate %d bytes\n", ARGV0,
				io[p].bs);
			terminate(io);
			return EX_OSERR;
		}else if(io[p].seek > 0) /* seek */
			switch(Io_seek(&io[p])){
			case EX_OK:	terminate(output(io)); return EX_OK;
			}
	}

	do{	/* read */
		Io_read(&io[0]);
		if(!noerror && io[0].bufuse == 0)
			Io_read(&io[0]); /* second chance */
		if(io[0].bufuse == 0) /* that's all she wrote */
			break;
		else if(io[0].bufuse < io[0].bs){
			io[0].prec += 1;
			if(noerror){
				output(io);
			}else
				count = 1;
			if(align >= 0){
				memset(io[0].buf + io[0].bufuse, align,
					io[0].bs - io[0].bufuse);
				io[0].bufuse = io[0].bs;
			}
		}

		/* write */ while(io[0].bufuse > 0){
			memcpy(io[1].buf, io[0].buf, io[1].bs);
			memmove(io[0].buf, io[0].buf + io[1].bs, io[0].bs - io[1].bs);

		}
	}while(count == 0 || --count > 0);

	terminate(output(io));
	
	return EX_OK;
}
